/* graphio.cc
 * Copyright (C) 2001 The University of Newcastle upon Tyne
 *
 * Written by Paul Coates (Paul.Coates@ncl.ac.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <sys/time.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <pwd.h>
#include "graphio.h"
#include "misc.h"

unsigned long init_block[4][2] =
{
	{300, 1209600},
	{1800, 4838400},
	{3600, 15724800},
	{10800, 62899200}
};

char *init_oid[] = { ".iso.3.6.1.2.1.2.2.1.10", ".iso.3.6.1.2.1.2.2.1.12", ".iso.3.6.1.2.1.2.2.1.13", ".iso.3.6.1.2.1.2.2.1.14", ".iso.3.6.1.2.1.2.2.1.16", ".iso.3.6.1.2.1.2.2.1.18", ".iso.3.6.1.2.1.2.2.1.19", ".iso.3.6.1.2.1.2.2.1.20" };

char *init_label[] = { "Octets In", "NonUnicast In", "Discards In", "Errors In", "Octets Out", "NonUnicast Out", "Discards Out", "Errors Out" };

int init_rate[8] = { 1, 1, 2, 2, 1, 1, 2, 2 };

int init_rgb[8][3] =
{
	{65535, 0, 0},
	{0, 65535, 0},
	{0, 0, 65535},
	{65535, 0, 65535},
	{0, 65535, 65535},
	{65535, 65535, 0},
	{65535, 32767, 0},
	{32767, 0, 65535}
};

GraphBlock::GraphBlock()
{
	gap = 0;
	howlong = 0;
	length = 0;
	start = 0;
	offset = 0;
	wrapped = 'n';
}

GraphType::GraphType()
{
	strcpy(oid, "");
	strcpy(label, "");
	rate = 1;
	r = g = b = 0;
}

GraphIO::GraphIO()
{
	char *md;
	struct passwd *pwd;

	md = getenv("NOMAD_DIR");
	if (md!=NULL) { strcpy(datadir, md); } else
	{ pwd = getpwuid(getuid()); sprintf(datadir, "%s/.nomad", pwd->pw_dir); }

	init_global();
}

GraphIO::~GraphIO()
{
}

void GraphIO::init_global()
{
	int i, first_start;

	num_global_blocks = 4;
	num_global_values = 8;

	first_start = 12 + (num_global_blocks*16) + (num_global_values*140);

	for (i=0; i<num_global_blocks; i++)
	{
		global_blocks[i].gap = init_block[i][0];
		global_blocks[i].howlong = init_block[i][1];
		if (global_blocks[i].gap>0)
			global_blocks[i].length =
				(global_blocks[i].howlong/global_blocks[i].gap)*
					(8+1+(8*num_global_values));
		global_blocks[i].start =
			(i==0)?first_start:(global_blocks[i-1].start+global_blocks[i-1].length);
	}
	for (i=0; i<num_global_values; i++)
	{
		strcpy(global_types[i].oid, init_oid[i]);
		strcpy(global_types[i].label, init_label[i]);
		global_types[i].rate = init_rate[i];
		global_types[i].r = init_rgb[i][0];
		global_types[i].g = init_rgb[i][1];
		global_types[i].b = init_rgb[i][2];
	}
}

void GraphIO::init_local()
{
	int i;

	num_blocks = num_global_blocks;
	num_values = num_global_values;
	for (i=0; i<num_blocks; i++)
	{
		blocks[i].gap     = global_blocks[i].gap;
		blocks[i].howlong = global_blocks[i].howlong;
		blocks[i].length  = global_blocks[i].length;
		blocks[i].start   = global_blocks[i].start;
		blocks[i].offset  = global_blocks[i].start;
		blocks[i].wrapped = 'n';
	}
	for (i=0; i<num_values; i++)
	{
		types[i].oid = global_types[i].oid;
	}
}

// save global settings to default location
int GraphIO::write_global()
{
	FILE *f;
	char str[256];
	int i, j;

	sprintf(str, "%s/graph.dat", datadir);
	f = fopen(str, "w");
	if (!f) return 0;

	strcpy(str, "NMD1"); for (i=0; i<4; i++) fputc(str[i], f);
	write_int32(f, num_global_blocks);
	write_int32(f, num_global_values);
	for (i=0; i<num_global_blocks; i++)
	{
		write_ulong32(f, global_blocks[i].gap);
		write_ulong32(f, global_blocks[i].howlong);
		write_ulong32(f, global_blocks[i].length);
		write_ulong32(f, global_blocks[i].start);
	}
	for (i=0; i<num_global_values; i++)
	{
		for (j=0; j<64; j++) fputc(global_types[i].oid[j], f);
		for (j=0; j<64; j++) fputc(global_types[i].label[j], f);
		write_int32(f, global_types[i].rate);
		write_int32(f, global_types[i].r);
		write_int32(f, global_types[i].g);
		write_int32(f, global_types[i].b);
	}
	fclose(f);

	sprintf(str, "%s/graph.oid", datadir);
	f = fopen(str, "w");
	if (!f) return 0;
	for (i=0; i<num_global_values; i++)
		fprintf(f, "%s\n", global_types[i].oid);
	fclose(f);

	return 1;
}

// load global settings from default location
int GraphIO::read_global()
{
	FILE *f;
	char str[256];
	int i, j;

	sprintf(str, "%s/graph.dat", datadir);
	f = fopen(str, "r");
	if (!f) return 0;

	fseek(f, 4L, SEEK_SET);
	num_global_blocks = read_int32(f);
	num_global_values = read_int32(f);
	for (i=0; i<num_global_blocks; i++)
	{
		global_blocks[i].gap     = read_ulong32(f);
		global_blocks[i].howlong = read_ulong32(f);
		global_blocks[i].length  = read_ulong32(f);
		global_blocks[i].start   = read_ulong32(f);
	}
	for (i=0; i<num_global_values; i++)
	{
		for (j=0; j<64; j++) global_types[i].oid[j] = fgetc(f);
		global_types[i].oid[64] = '\0';
		for (j=0; j<64; j++) global_types[i].label[j] = fgetc(f);
		global_types[i].label[64] = '\0';
		global_types[i].rate = read_int32(f);
		global_types[i].r = read_int32(f);
		global_types[i].g = read_int32(f);
		global_types[i].b = read_int32(f);
	}
	fclose(f);

	return 1;
}

// compare global and local settings to make sure they are in sync
int GraphIO::valid()
{
	int i;

	if (num_global_blocks!=num_blocks) return 0;
	if (num_global_values!=num_values) return 0;
	for (i=0; i<num_blocks; i++)
	{
		if (global_blocks[i].gap     != blocks[i].gap) return 0;
		if (global_blocks[i].howlong != blocks[i].howlong) return 0;
		if (global_blocks[i].length  != blocks[i].length) return 0;
		if (global_blocks[i].start   != blocks[i].start) return 0;
	}
	for (i=0; i<num_values; i++)
	{
		if (strcmp(global_types[i].oid, types[i].oid)!=0) return 0;
	}

	return 1;
}

// upgrade the file to match the global settings
int GraphIO::update(char *fname)
{
	return 1;
}

void GraphIO::load_header(FILE *f)
{
	int i, j;

	fseek(f, 4L, SEEK_SET);
	num_blocks = read_int32(f);
	num_values = read_int32(f);
	for (i=0; i<num_blocks; i++)
	{
		blocks[i].gap     = read_ulong32(f);
		blocks[i].howlong = read_ulong32(f);
		blocks[i].length  = read_ulong32(f);
		blocks[i].start   = read_ulong32(f);
		blocks[i].offset  = read_ulong32(f);
		blocks[i].wrapped = fgetc(f);
	}
	for (i=0; i<num_values; i++)
	{
		for (j=0; j<64; j++) types[i].oid[j] = fgetc(f);
		types[i].oid[64] = '\0';
	}
}

// create a new file using global settings
int GraphIO::create(char *fname)
{
	FILE *f;
	char buf[8];
	int i, j, count;

	init_local();

	f = fopen(fname, "w");
	if (!f) return 0;

	count = 0;
	strcpy(buf, "NMD1"); for (i=0; i<4; i++) fputc(buf[i], f);
	write_int32(f, num_blocks);
	write_int32(f, num_values);
	for (i=0; i<num_blocks; i++)
	{
		write_ulong32(f, blocks[i].gap);
		write_ulong32(f, blocks[i].howlong);
		write_ulong32(f, blocks[i].length);
		write_ulong32(f, blocks[i].start);
		write_ulong32(f, blocks[i].offset);
		fputc((int)blocks[i].wrapped, f);
		count += blocks[i].length;
	}
	for (i=0; i<num_values; i++)
	{
		for (j=0; j<64; j++) fputc(types[i].oid[j], f);
	}
	for (i=0; i<count; i++) fputc(0, f);
	fclose(f);

	return 1;
}

// add local data to file
int GraphIO::add(char *fname)
{
	FILE *f;

	f = fopen(fname, "r");
	if (f) fclose(f); else
	{
		if (create(fname)==0) return 0;
	}

	f = fopen(fname, "r+");
	if (f)
	{
		load_header(f);
		add_recursive(f, 0);
		fclose(f);
	}

	return 1;
}

void GraphIO::add_recursive(FILE *f, int lvl)
{
	unsigned long tmpvals[MAX_VALUES];
	TimeStamp tmptime;
	char tmpflag;
	int i, tsl;

	// if removing last entry throw it away
	if (lvl>=num_blocks) return;

	// if too soon for this buffer throw it away
	// grab timestamp of previous record (tmptime)
	if (blocks[lvl].offset - 8+1+(8*num_values) < blocks[lvl].start)
		tsl = blocks[lvl].start + blocks[lvl].length - 8+1+(8*num_values);
	else
		tsl = blocks[lvl].offset - 8+1+(8*num_values);

	fseek(f, tsl, SEEK_SET);
	tmptime.read_time64(f);
	if ((unsigned int)(current.compare(tmptime))<blocks[lvl].gap) return;


// REPLACE THE FOLLOWING BIT WITH (create new average by agregating several records and creating a new average, e.g. lvl==0 so we are grabbing 5 minute values, if we were to copy over data to the next block it would be half hour values so we need to copy the oldest 6 values (5*6 -> 30) and record the average. This smooths out the graphs correctly for historical data.

	// copy original record
	fseek(f, blocks[lvl].offset, SEEK_SET);
	tmptime.read_time64(f);
	tmpflag = fgetc(f);
	for (i=0; i<num_values; i++)
		tmpvals[i] = read_ulong64(f);

	// write record to blocks[lvl].offset
	fseek(f, blocks[lvl].offset, SEEK_SET);
	current.write_time64(f);
	fputc(status, f);
	for (i=0; i<num_values; i++)
		write_ulong64(f, value[i]);

	// advance the offset
	blocks[lvl].offset += 8+1+(8*num_values);

	if (blocks[lvl].wrapped == 'n')
	{
		if (blocks[lvl].offset >= blocks[lvl].start + blocks[lvl].length)
		{
			blocks[lvl].offset = blocks[lvl].start;
			blocks[lvl].wrapped = 'y';
		}
	}else
	{
		if (blocks[lvl].offset >= blocks[lvl].start + blocks[lvl].length)
		{
			blocks[lvl].offset = blocks[lvl].start;
		}
		// pass record on to next buffer
		current = tmptime;
		status = tmpflag;
		for (i=0; i<num_values; i++)
			value[i] = tmpvals[i];
		add_recursive(f, lvl+1);
	}
	// write back header stuff
	fseek(f, 12+(21*lvl)+16, SEEK_SET);
	write_ulong32(f, blocks[lvl].offset);
	fputc((int)blocks[lvl].wrapped, f);
}

// get data from file and store in DataSet ready for graph drawing code
int GraphIO::get(char *fname, DataSet *d)
{
	int i, j, len, count, recsize;
	FILE *f;

//	unsigned long diff;
//	struct timeval st, et;
//	gettimeofday(&st, NULL);

	f = fopen(fname, "r");
	if (!f) return 0;

	load_header(f);

	// work out size for data structures
	count = 0;
	for (i=0; i<num_blocks; i++)
	{
		count += (blocks[i].howlong/blocks[i].gap);
	}
	d->num_variables = num_values;
	d->num_entries = count;

	// free up existing memory allocation
	if (d->timestamps != NULL) delete [] d->timestamps;
	if (d->variables != NULL) delete []  d->variables;
	if (d->status != NULL) delete []  d->status;

	// allocate memory for data structures
	d->timestamps = new TimeStamp[count];
	d->status = new char[count];
	d->variables = new DataLine[num_values];
	for (i=0; i<num_values; i++)
	{
		d->variables[i].values = new unsigned long[count];
	}

//	gettimeofday(&et, NULL);
//	diff = (1000000*(et.tv_sec-st.tv_sec))+(et.tv_usec-st.tv_usec);
//	printf("GraphIO::get(3) : %lu\n", diff);

	count = 0;
	recsize = 8+1+(8*num_values);
	for (i=num_blocks-1; i>=0; i--)
	{
		if (blocks[i].wrapped=='y')
		{
			len = (blocks[i].start+blocks[i].length-blocks[i].offset)/recsize;
			fseek(f, blocks[i].offset, SEEK_SET);
			for (j=0; j<len; j++)
			{
				get_record(d, f, count);
				count++;
			}
		}
		len = (blocks[i].offset-blocks[i].start)/recsize;
		fseek(f, blocks[i].start, SEEK_SET);
		for (j=0; j<len; j++)
		{
			get_record(d, f, count);
			count++;
		}

//		gettimeofday(&et, NULL);
//		diff = (1000000*(et.tv_sec-st.tv_sec))+(et.tv_usec-st.tv_usec);
//		printf("GraphIO::get(block %d) : %lu\n", i, diff);
	}
	d->num_entries = count;

	fclose(f);


//	gettimeofday(&et, NULL);
//	diff = (1000000*(et.tv_sec-st.tv_sec))+(et.tv_usec-st.tv_usec);
//	printf("GraphIO::get(end) : %lu\n", diff);

	return 1;
}

void GraphIO::get_record(DataSet *d, FILE *f, int count)
{
	int j;

	d->timestamps[count].read_time64(f);
	d->status[count] = fgetc(f);
	for (j=0; j<num_values; j++)
		d->variables[j].values[count] = read_ulong64(f);
}

